/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://treleis.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.binding.soap;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.xml.XMLConstants;
import javax.xml.namespace.QName;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;
import javax.xml.validation.Validator;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.Resource;
import org.bluestemsoftware.specification.eoa.component.ComponentContext;
import org.bluestemsoftware.specification.eoa.component.ComponentDeployment;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentName;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentType;
import org.bluestemsoftware.specification.eoa.component.binding.Binding;
import org.bluestemsoftware.specification.eoa.component.binding.rt.BindingRT;
import org.bluestemsoftware.specification.eoa.ext.binding.BindingFactory;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAP11HTTPBindingRT;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAP11VMBindingRT;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAP12HTTPBindingRT;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAP12VMBindingRT;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAPBindingConfiguration;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAPBindingRT;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.addressing.soap.WSAFeature;
import org.bluestemsoftware.specification.eoa.ext.wsdl.WSDLDocument;
import org.bluestemsoftware.specification.eoa.ext.wsdl.WSDLException;
import org.bluestemsoftware.specification.eoa.ext.wsdl.wsdl11.WSDL11Document;
import org.bluestemsoftware.specification.eoa.ext.wsdl.wsdl20.ExtendedDocument;
import org.bluestemsoftware.specification.eoa.ext.wsdl.wsdl20.WSDL20Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/**
 * A type specific <code>Binding</code> factory that creates bindings which conform to SOAP
 * specification.
 */
public class SOAPBindingFactory extends BindingFactory {

    /*
     * (non-Javadoc)
     * 
     * SPI
     */
    public interface Provider extends BindingFactory.Provider {

        public SOAPBindingRT.Provider spi_createDefaultProvider(SOAPBindingFactory factory, ComponentName componentName, String protocolType) throws DeploymentException;

        public SOAPBindingRT.Provider spi_createProvider(SOAPBindingFactory factory, ComponentName componentName, String protocolType, Set<String> features, SOAPBindingConfiguration sbc) throws DeploymentException;

    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.ExtensionFactory#getExtensionType()
     */
    public String getExtensionType() {
        return SOAPBinding.TYPE;
    }
    
    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.binding.BindingFactory#getBindingType()
     */
    @Override
    public String getBindingType() {
        return SOAPBinding.BINDING_TYPE;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.binding.BindingFactory#createDefaultProvider(org.bluestemsoftware.specification.eoa.component.ComponentContext,
     *      org.bluestemsoftware.specification.eoa.component.binding.Binding)
     */
    @Override
    public SOAPBindingRT createDefaultBinding(ComponentContext componentContext, Binding runtimeProvidable) throws DeploymentException {

        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();

        try {

            thread.setContextClassLoader(factoryContext.getClassLoader());

            if (!(runtimeProvidable instanceof SOAPBinding)) {
                throw new DeploymentException("Expected instanceof " + SOAPBinding.class.getName());
            }

            ComponentName componentName = runtimeProvidable.getName();
            String pt = runtimeProvidable.getUnderlyingProtocol().getProtocolType();
            SOAPBindingRT.Provider p = ((Provider)provider).spi_createDefaultProvider(this, componentName, pt);
            
            if (pt.equals(SOAP11HTTPProtocol.TYPE)) {
                return new SOAP11HTTPBindingRT(this, (SOAP11HTTPBindingRT.Provider)p, componentName, pt);
            } else if (pt.equals(SOAP12HTTPProtocol.TYPE)) {
                return new SOAP12HTTPBindingRT(this, (SOAP12HTTPBindingRT.Provider)p, componentName, pt);
            } else if (pt.equals(SOAP11VMTransportProtocol.TYPE)) {
                return new SOAP11VMBindingRT(this, (SOAP11VMBindingRT.Provider)p, componentName, pt);
            } else if (pt.equals(SOAP12VMTransportProtocol.TYPE)) {
                return new SOAP12VMBindingRT(this, (SOAP12VMBindingRT.Provider)p, componentName, pt);
            } else {
                throw new DeploymentException("Error creating default provider definition "
                        + componentName
                        + ". Corresponding Binding employs an unsupported protocol "
                        + pt
                        + ".");
            }

        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.wsdl.WSDLFactory$BindingDefinitionReader#readBinding(org.bluestemsoftware.specification.eoa.component.ComponentContext,
     *      java.lang.String, org.w3c.dom.Element,
     *      org.bluestemsoftware.specification.eoa.ext.wsdl.WSDLFactory.BindingDefinitionReader.BindingStyle)
     */
    public void readBinding(ComponentContext componentContext, String targetNamespace, Element bindingElement, BindingStyle bindingStyle) throws DeploymentException {

        if (factoryContext == null) {
            throw new IllegalStateException("Factory has no context.");
        }
        if (componentContext == null) {
            throw new IllegalArgumentException("componentContext null");
        }
        if (targetNamespace == null) {
            throw new IllegalArgumentException("targetNamespace null");
        }
        if (bindingElement == null) {
            throw new IllegalArgumentException("bindingElement null");
        }

        Binding binding = null;
        if (bindingElement.getNamespaceURI().equals(WSDL20Document.TYPE)) {
            binding = org.bluestemsoftware.specification.eoa.ext.wsdl.wsdl20.ext.SOAPBindingReader.readBinding(
                    this, componentContext, targetNamespace,
                    (org.bluestemsoftware.specification.eoa.ext.wsdl.wsdl20.BindingElement)bindingElement,
                    bindingStyle);
        } else if (bindingElement.getNamespaceURI().equals(WSDL11Document.TYPE)) {
            binding = org.bluestemsoftware.specification.eoa.ext.wsdl.wsdl11.ext.SOAPBindingReader.readBinding(
                    this, componentContext, targetNamespace,
                    (org.bluestemsoftware.specification.eoa.ext.wsdl.wsdl11.BindingElement)bindingElement,
                    bindingStyle);
        } else {
            throw new DeploymentException("Unsupported wsdl version " + bindingElement.getNamespaceURI() + ".");
        }

        componentContext.getDeployment().addComponent(binding);

    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.wsdl.WSDLFactory$BindingWriter#writeBinding(org.bluestemsoftware.specification.eoa.ext.wsdl.WSDLDocument,
     *      org.bluestemsoftware.specification.eoa.component.binding.Binding)
     */
    public void writeBinding(WSDLDocument wsdlDocument, Binding binding) throws WSDLException {
        if (wsdlDocument == null) {
            throw new IllegalArgumentException("wsdlDocument null");
        }
        if (binding == null) {
            throw new IllegalArgumentException("binding null");
        }
        if (wsdlDocument instanceof WSDL11Document) {
            org.bluestemsoftware.specification.eoa.ext.wsdl.wsdl11.ext.SOAPBindingWriter.writeBinding(
                    (WSDL11Document)wsdlDocument, (SOAPBinding)binding);
        } else if (wsdlDocument instanceof ExtendedDocument) {
            org.bluestemsoftware.specification.eoa.ext.wsdl.wsdl20.ext.SOAPBindingWriter.writeBinding(
                    (ExtendedDocument)wsdlDocument, (SOAPBinding)binding);
        } else {
            throw new WSDLException("Unsupported wsdlDocument type " + wsdlDocument.getClass().getName() + ".");
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.ProviderReader#readProviders(org.bluestemsoftware.specification.eoa.component.ComponentContext,
     *      javax.xml.transform.dom.DOMSource)
     */
    public Set<BindingRT> readProviders(ComponentContext componentContext, DOMSource source) throws DeploymentException {

        if (componentContext == null) {
            throw new IllegalArgumentException("entityResolver null");
        }
        if (source == null) {
            throw new IllegalArgumentException("source null");
        }

        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {

            thread.setContextClassLoader(factoryContext.getClassLoader());

            validateDefinition(source);
            
            ComponentDeployment deployment = componentContext.getDeployment();

            Element definition = (Element)source.getNode();
            String prefixedName = definition.getAttribute("name");
            ComponentName cname = null;
            if (prefixedName.equals("")) {
                throw new DeploymentException("Error loading runtime provider definition "
                        + ". Definition missing required attribute 'name'.");
            } else {
                cname = ComponentName.valueOf(definition, definition.getAttribute("name"));
            }

            Binding binding = (Binding)deployment.getComponent(ComponentType.BINDING, cname);

            if (binding == null) {
                throw new DeploymentException("Error loading runtime provider definition "
                        + cname
                        + ". Definition for corresponding RootComponent is undefined.");
            }

            if (!(binding instanceof SOAPBinding)) {
                throw new DeploymentException("Error loading runtime provider definition "
                        + cname
                        + ". Corresponding RootComponent is not an instance of "
                        + SOAPBinding.class);
            }
            
            Set<String> features = new HashSet<String>();
            features.add(WSAFeature.TYPE);
            QName childName = new QName(SOAPBinding.TYPE, "features");

            Element featuresElement = getChildElement(definition, childName);
            if (featuresElement != null) {
                childName = new QName(SOAPBinding.TYPE, "feature");
                List<Element> featureElements = getChildElements(featuresElement, childName);
                for (Element featureElement : featureElements) {
                    features.add(getText(featureElement));
                }
            }
            
            SOAPBindingConfiguration sbc = new SOAPBindingConfiguration();

            childName = new QName(SOAPBinding.TYPE, "configuration");
            Element configurationElement = getChildElement(definition, childName);

            if (configurationElement != null) {
                
                childName = new QName(SOAPBinding.TYPE, "operations");
                Element operationsElement = getChildElement(configurationElement, childName);
                if (operationsElement != null) {
                    if (binding.isReusable()) {
                        throw new DeploymentException("Error loading runtime provider definition "
                                + cname
                                + ". Operations element defined within configuration, but binding"
                                + " is reusable.");
                    }
                    childName = new QName(SOAPBinding.TYPE, "operation");
                    Map<QName, Integer> requestorTimeout = new HashMap<QName, Integer>();
                    List<Element> list = getChildElements(operationsElement, childName);
                    for (Element operationElement : list) {
                        prefixedName = operationElement.getAttribute("name");
                        QName operationName = ComponentName.valueOf(operationElement, prefixedName);
                        if (binding.getOperation(operationName) == null) {
                            throw new DeploymentException("Error loading runtime provider definition "
                                    + cname
                                    + ". Operation element defined within configuration for operation "
                                    + operationName
                                    + ", but corresponding binding defines no such operation.");
                        }
                        childName = new QName(SOAPBinding.TYPE, "requestorTimeout");
                        Element timeoutElement = getChildElement(operationElement, childName);
                        Integer timeout = Integer.valueOf(getText(timeoutElement));
                        requestorTimeout.put(operationName, timeout);
                    }
                    sbc.setRequestorTimeout(requestorTimeout);
                }
                
                childName = new QName(SOAPBinding.TYPE, "dataSourceConnector");
                Element dsElement = getChildElement(configurationElement, childName);
                if (dsElement != null) {
                    sbc.setDataSourceConnectorName(getText(dsElement));
                }

                childName = new QName(SOAPBinding.TYPE, "embedStackTraceInFaults");
                Element embedElement = getChildElement(configurationElement, childName);
                if (embedElement != null) {
                    sbc.setEmbeddStackTraceInFaults(Boolean.parseBoolean(getText(embedElement)));
                }

                childName = new QName(SOAPBinding.TYPE, "largeAttachmentSize");
                Element largeAttachmentElement = getChildElement(configurationElement, childName);
                if (largeAttachmentElement != null) {
                    sbc.setLargeAttachmentSize(Integer.parseInt(getText(largeAttachmentElement)));
                }

                childName = new QName(SOAPBinding.TYPE, "requestorTimeout");
                Element anonymousElement = getChildElement(configurationElement, childName);
                if (anonymousElement != null) {
                    sbc.setDefaultRequestorTimeout(Integer.parseInt(getText(anonymousElement)));
                }

                childName = new QName(SOAPBinding.TYPE, "maxPayloadProcessors");
                Element payloadElement = getChildElement(configurationElement, childName);
                if (payloadElement != null) {
                    int value = Integer.parseInt(getText(payloadElement));
                    sbc.setMaxPayloadProcessors(value);
                }

                childName = new QName(SOAPBinding.TYPE, "minHeaderProcessors");
                Element headersElement = getChildElement(configurationElement, childName);
                if (headersElement != null) {
                    int value = Integer.parseInt(getText(headersElement));
                    sbc.setMinHeaderProcessors(value);
                }
            }

            String pt = binding.getUnderlyingProtocol().getProtocolType();
            Set<BindingRT> providers = new HashSet<BindingRT>();

            SOAPBindingRT bindingRT = null;
            if (pt.equals(SOAP11HTTPProtocol.TYPE)) {
                SOAPBindingRT.Provider p = ((Provider)provider).spi_createProvider(this, cname, pt, features, sbc);
                bindingRT = new SOAP11HTTPBindingRT(this, (SOAP11HTTPBindingRT.Provider)p, cname, pt);
            } else if (pt.equals(SOAP12HTTPProtocol.TYPE)) {
                SOAPBindingRT.Provider p = ((Provider)provider).spi_createProvider(this, cname, pt, features, sbc);
                bindingRT = new SOAP12HTTPBindingRT(this, (SOAP12HTTPBindingRT.Provider)p, cname, pt);
            } else if (pt.equals(SOAP11VMTransportProtocol.TYPE)) {
                SOAPBindingRT.Provider p = ((Provider)provider).spi_createProvider(this, cname, pt, features, sbc);
                bindingRT = new SOAP11VMBindingRT(this, (SOAP11VMBindingRT.Provider)p, cname, pt);
            } else if (pt.equals(SOAP12VMTransportProtocol.TYPE)) {
                SOAPBindingRT.Provider p = ((Provider)provider).spi_createProvider(this, cname, pt, features, sbc);
                bindingRT = new SOAP12VMBindingRT(this, (SOAP12VMBindingRT.Provider)p, cname, pt);
            } else {
                throw new DeploymentException("Error creating provider definition "
                        + cname
                        + ". Corresponding Binding employs an unsupported protocol "
                        + pt
                        + ".");
            }

            providers.add(bindingRT);
            return providers;

        } finally {
            thread.setContextClassLoader(cl);
        }

    }
    
    private Element getChildElement(Element parent, QName childName) {
        NodeList nodeList = parent.getChildNodes();
        for (int i = 0; i < nodeList.getLength(); i++) {
            if (nodeList.item(i).getNodeType() != Node.ELEMENT_NODE) {
                continue;
            }
            Element element = (Element)nodeList.item(i);
            if (childName.getNamespaceURI().equals(XMLConstants.NULL_NS_URI)) {
                if (element.getNamespaceURI() != null) {
                    continue;
                }
            } else {
                if (element.getNamespaceURI() == null) {
                    continue;
                }
                if (!childName.getNamespaceURI().equals("*")
                        && !element.getNamespaceURI().equals(childName.getNamespaceURI())) {
                    continue;
                }
            }
            if (!childName.getLocalPart().equals("*") && !element.getLocalName().equals(childName.getLocalPart())) {
                continue;
            }
            return element;
        }
        return null;
    }

    private List<Element> getChildElements(Element parent, QName childName) {
        List<Element> answer = new ArrayList<Element>();
        NodeList nodeList = parent.getChildNodes();
        for (int i = 0; i < nodeList.getLength(); i++) {
            if (nodeList.item(i).getNodeType() != Node.ELEMENT_NODE) {
                continue;
            }
            Element element = (Element)nodeList.item(i);
            if (childName.getNamespaceURI().equals(XMLConstants.NULL_NS_URI)) {
                if (element.getNamespaceURI() != null) {
                    continue;
                }
            } else {
                if (element.getNamespaceURI() == null) {
                    continue;
                }
                if (!childName.getNamespaceURI().equals("*")
                        && !element.getNamespaceURI().equals(childName.getNamespaceURI())) {
                    continue;
                }
            }
            if (!childName.getLocalPart().equals("*") && !element.getLocalName().equals(childName.getLocalPart())) {
                continue;
            }
            answer.add(element);
        }
        return answer;
    }

    private void validateDefinition(DOMSource source) throws DeploymentException {

        Resource resource = null;
        try {
            resource = factoryContext.getResource(SOAPBinding.PROVIDER_SCHEMA_LOCATION);
        } catch (DeploymentException de) {
            throw new DeploymentException("Error validating definition. " + de);
        }

        SchemaFactory schemaFactory = SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);

        javax.xml.validation.Schema schema = null;
        try {
            schema = schemaFactory.newSchema(new StreamSource(resource.getInputStream()));
        } catch (Exception ex) {
            throw new DeploymentException("Error parsing schema. " + ex);
        }
        DOMValidator domValidator = new DOMValidator(schema);
        String validationError;
        try {
            validationError = domValidator.validate(source);
        } catch (Exception ex) {
            throw new DeploymentException("Error validating definition. " + ex);
        }
        if (validationError != null) {
            throw new DeploymentException("Engine definition invalid. " + validationError);
        }

    }
    
    private String getText(Element element) {
        if (element == null || !element.hasChildNodes()) {
            return null;
        }
        try {
            Text textNode = (Text)element.getFirstChild();
            return textNode.getData();
        } catch (ClassCastException ce) {
        }
        return null;
    }

    static final class DOMValidator {

        private Schema schema;

        public DOMValidator(Schema schema) {
            this.schema = schema;
        }

        public String validate(DOMSource source) throws SAXException, IOException {
            Validator validator = schema.newValidator();
            validator.setErrorHandler(new ErrorHandlerImpl());
            validator.validate(source);
            return ((ErrorHandlerImpl)validator.getErrorHandler()).getErrors();
        }

        static class ErrorHandlerImpl implements ErrorHandler {

            private ArrayList<String> errorMessages = new ArrayList<String>();

            public void error(SAXParseException spe) throws SAXException {
                errorMessages.add(spe.getMessage());
            }

            public void warning(SAXParseException spe) throws SAXException {
                errorMessages.add(spe.getMessage());
            }

            public void fatalError(SAXParseException spe) throws SAXException {
                errorMessages.add(spe.getMessage());
            }

            public String getErrors() {
                if (errorMessages.size() > 0) {
                    Iterator<String> messageIterator = errorMessages.iterator();
                    StringBuilder errorMessage = new StringBuilder();
                    while (messageIterator.hasNext()) {
                        errorMessage.append(messageIterator.next() + "  ");
                    }
                    return errorMessage.toString();
                }
                return null;
            }

        }

    }


}
