/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://treleis.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.binding.soap.rt;

import java.util.Map;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.ext.connector.db.DataSourceConnector;

/**
 * Defines configuration used to configure a runtime <code>SOAPBinding</code> component
 * provider.
 */
public class SOAPBindingConfiguration {

    // initialize all instance vars with reasonable defaults

    private boolean embedStackTraceInFaults = true;
    private int largeAttachmentSize = 0; // do not buffer
    private int defaultRequestorTimeout = 60000; // 1 minute
    private Map<QName, Integer> requestorTimeouts;
    private int maxPayloadProcessors = 5;
    private int minHeaderProcessors = 5;
    private String dataSourceConnectorName = DataSourceConnector.SYSTEM_DATA_SOURCE;
    
    /**
     * Defines DataSourceConnector name which MAY be required by binding provider to implement
     * message and/or header persistence. If element is undefined, the connector
     * 'SystemDataSource' is implied.
     * 
     * @return
     */
    public String getDataSourceConnectorName() {
        return dataSourceConnectorName;
    }
    
    public void setDataSourceConnectorName(String dataSourceConnectorName) {
        this.dataSourceConnectorName = dataSourceConnectorName;
    }
    
    /**
     * Gets amount of time before a blocking request times out when no timeout is specified for
     * a specific operation, i.e the amount of time provider will spend waiting on an anonymous
     * response before throwing a fault.
     * 
     * @return value in milliseconds or -1 to indicate no timeout.
     */
    public int getDefaultRequestorTimeout() {
        return defaultRequestorTimeout;
    }

    public void setDefaultRequestorTimeout(int defaultRequestorTimeout) {
        this.defaultRequestorTimeout = defaultRequestorTimeout;
    }

    /**
     * Gets amount of time before a blocking request times out for a specific operation, i.e
     * the amount of time provider will spend waiting on an anonymous response before throwing
     * a fault.
     * 
     * @return value in milliseconds or <code>null</code> value is undefined for indicated
     * operation. A value of -1 indicates no timeout.
     */
    public Integer getRequestorTimeout(QName operationName) {
        if (requestorTimeouts != null) {
            return requestorTimeouts.get(operationName);
        }
        return null;
    }

    public void setRequestorTimeout(Map<QName, Integer> requestorTimeouts) {
        this.requestorTimeouts = requestorTimeouts;
    }

    /**
     * Defines maximum number of concurrent processors which process message payload, i.e.
     * which process action by binding it according to wsdl description and then passing to
     * engine.
     * 
     * @return
     */
    public int getMaxPayloadProcessors() {
        return maxPayloadProcessors;
    }

    public void setMaxPayloadProcessors(int maxPayloadProcessors) {
        this.maxPayloadProcessors = maxPayloadProcessors;
    }

    /**
     * Defines minimum number of processors which process message headers and write message to
     * queue for payload processing. Note that maximum number of processors is unconstrained,
     * i.e. to facillitate loose coupling. Idle processors are reclaimed as needed.
     * 
     * @return
     */
    public int getMinHeaderProcessors() {
        return minHeaderProcessors;
    }

    public void setMinHeaderProcessors(int minHeaderProcessors) {
        this.minHeaderProcessors = minHeaderProcessors;
    }

    /**
     * Gets value which MUST be used by provider to limit the size of message attachments
     * buffered to memory.
     * 
     * @return size in bytes or -1 to indicate size is unconstrained. Note that if size is
     *         unconstrained, provider SHOULD be optimized to handle large attachments.
     */
    public int getLargeAttachmentSize() {
        return largeAttachmentSize;
    }

    public void setLargeAttachmentSize(int largeAttachmentSize) {
        this.largeAttachmentSize = largeAttachmentSize;
    }

    /**
     * Constrains fault handling.
     * 
     * @return true if stack trace embedded in detail. false if not.
     */
    public boolean isEmbeddStackTraceInFaults() {
        return embedStackTraceInFaults;
    }

    public void setEmbeddStackTraceInFaults(boolean embedStackTraceInFaults) {
        this.embedStackTraceInFaults = embedStackTraceInFaults;
    }

}
